#define _BSD_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#ifndef N
#define N 4000
#endif /* N */

#ifndef STRIDE
#define STRIDE (N)
#endif /* STRIDE */

#define SIZE (N * STRIDE)


void init_enumerate(float * m, size_t stride, size_t width, size_t height) {
    for (size_t y = 0; y < N; ++y) {
        for (size_t x = 0; x < N; ++x) {
            m[x + y * STRIDE] = x + y * width;
        }
        for (size_t x = N; x < STRIDE; ++x) {
            m[x + y * STRIDE] = - 1.0f;
        }
    }
}


void transpose_2d_simple(const float * __restrict__ in, size_t in_stride,
                         size_t width, size_t height,
                         float * __restrict__ out, size_t out_stride) {
    for (size_t y = 0; y < height; ++y) {
        for (size_t x = 0; x < width; ++x) {
            out[y + x * out_stride] = in[x + y * in_stride];
        }
    }
}


void check_transpose(const float * m, size_t stride, size_t width, size_t height) {
    for (size_t y = 0; y < height; ++y) {
        for (size_t x = 0; x < width; ++x) {
            if (m[x + y * stride] != (float) (y + x * height)) {
                printf ("(%lu,%lu) failed\n", y, x);
            }
        }
    }
}


int main(int argc, char ** argv) {
    // Allocate matrices
    float * a = malloc(SIZE * sizeof(float));
    float * b = malloc(SIZE * sizeof(float));

    // Init
    init_enumerate(a, STRIDE, N, N);
    memset(b, 0, SIZE * sizeof(float));

    // Start the clock
    struct timeval start, end, elapsed;
    gettimeofday(&start, NULL);

    // Transpose
    transpose_2d_simple(a, STRIDE, N, N, b, STRIDE);

    // Calculate elapsed time
    gettimeofday(&end, NULL);
    timersub(&end, &start, &elapsed);
    double time = elapsed.tv_sec + elapsed.tv_usec / 1000000.0;
    printf("Simple transpose time: %.4fs\n", time);

    // Check result
    check_transpose(b, STRIDE, N, N);

    // Clean up
    free(a);
    free(b);

    return 0;
}
